package org.bdware.server.action;

import com.google.gson.reflect.TypeToken;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.nodecenter.CMNode;
import org.bdware.server.nodecenter.NodeCenterActions;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

public class DistributeCallback extends ResultCallback {
    private static final Logger LOGGER = LogManager.getLogger(DistributeCallback.class);
    String sponsorPubkey;// 发起节点的
    Map<String, String> nodes;
    String pubKey; //发起节点的用户公钥
    String signature; // 发起节点的
    int count; // 集群中几个节点已成功启动
    ResultCallback res; // 返回给Node
    String fileName; // ypk的文件名
    int index;
    String distributeID;  //发起节点的分发id

    public DistributeCallback(String ss, String sponID, Map<String, String> n, ResultCallback re, String p, String s) {
        distributeID = ss;
        count = 0;
        index = 1;
        this.res = re;
        this.pubKey = p;
        this.signature = p;
        this.sponsorPubkey = sponID;
        this.nodes = n;
    }

    // 向集群中没有合约项目的节点发送合约
    public void distributeContractProject(String filePath, String isPrivate) {
        LOGGER.debug("[DistributeCallback] distributeContractProject : position----5" + filePath);

        File f = new File(filePath);
        fileName = f.getName();
        LOGGER.debug("[DistributeCallback] distributeContractProject : fileName=" + fileName);
        LOGGER.debug("[DistributeCallback] nodeNames: " + JsonUtil.toJson(nodes));

        for (String ID : nodes.keySet()) {
            if (ID.equals(sponsorPubkey))
                continue;
            CMNode node = NodeCenterActions.nodeInfos.get(ID);
            node.connection.sendProject(filePath, isPrivate, pubKey, node.ipPort, this);
        }
    }

    public void NCreceive(String progress) {
        Map<String, String> ret2 = new HashMap<>();
        ret2.put("action", "onDistributeContract");
        ret2.put("progress", "NC is receiving ypk from sponsor,progress is " + progress + "%.");

        Map<String, String> map = new HashMap<>();
        map.put("action", "onDistribute");
        map.put("content", JsonUtil.toJson(ret2));
        map.put("distributeID", distributeID);
        res.onResult(JsonUtil.toJson(map));
    }

    public void onDistribute(String progress, String nodeIP) {
        Map<String, String> ret2 = new HashMap<>();
        ret2.put("action", "onDistributeContract");
        ret2.put(
                "progress",
                "NC is sending ypk to NO."
                        + index
                        + " node in the units,progress is "
                        + progress
                        + "%.");
        if (progress.equals("100.00")) index++;

        Map<String, String> map = new HashMap<>();
        map.put("action", "onDistributeYPK");
        map.put("responseID", distributeID);
        map.put("content", JsonUtil.toJson(ret2));
        map.put("nodeIP", nodeIP);
        map.put("progress", progress);
        res.onResult(JsonUtil.toJson(map));
    }

    public void onReceive(Map<String, String> map) {
        LOGGER.debug("[DistributeCallback] onReceive :  position----9");
        String progress = map.get("progress");
        String nodeIP = map.get("ipPort");
        if (progress.equals("100")) {
            Map<String, String> args = new HashMap<>();
            args.put("fileName", fileName);
            args.put("pubKey", pubKey);
            args.put("signature", signature);
            count++;
            LOGGER.debug(count + "个节点已收完成" + "   总共有" + nodes.size() + "  个节点");
            if (count == nodes.size()) {
                // res返回给前端，合约分发完成
                Map<String, String> ret2 = new HashMap<>();
                ret2.put("action", "onDistributeFinish");
                ret2.put("progress", "100%");
                Map<String, String> map_send = new HashMap<>();
                map_send.put("action", "onDistribute");
                map_send.put("responseID", distributeID);
                map_send.put("over", "true");
                map_send.put("nodeIP", nodeIP);
                map_send.put("content", JsonUtil.toJson(ret2));
                res.onResult(JsonUtil.toJson(map_send));
                //NC delete file
                File file = new File("./temp/stateFiles/" + fileName);
                if (file.exists() && file.isFile())
                    file.delete();
            }
        }
    }

    @Override
    public void onResult(String str) {
        LOGGER.debug("[DistributeCallback] str=" + str);

        Map<String, String> map = JsonUtil.fromJson(str, new TypeToken<Map<String, String>>() {
        }.getType());

        NodeCenterActions.sync.sleepWithTimeout(map.get("requestID"), this, 60);

        String operation = map.get("operation");

        switch (operation) {
            case "NCreceive":
                NCreceive(map.get("progress"));
                break;
            case "distribute":
                distributeContractProject(map.get("receiveFileName"), map.get("isPrivate"));
                break;
            case "onDistribute":
                onDistribute(map.get("progress"), map.get("nodeIP"));
                break;
            case "onReceive":
                onReceive(map);
                break;
            default:
                LOGGER.debug("[DistributeCallback] undefined operation " + operation);
        }
    }
}
