package org.bdware.server.action;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractResult;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.nodecenter.CMNode;
import org.bdware.server.nodecenter.ContractExecutor;
import org.bdware.server.nodecenter.MultiPointContractInfo;
import org.bdware.server.nodecenter.NodeCenterActions;

import java.util.concurrent.atomic.AtomicInteger;

public class ResponseOnceExecutor implements ContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(ResponseOnceExecutor.class);
    MultiPointContractInfo info;
    AtomicInteger order = new AtomicInteger(0);

    public ResponseOnceExecutor(MultiPointContractInfo minfo) {
        info = minfo;
    }

    @Override
    public void execute(String requestID, ResultCallback rc, String req) {
        executeInternal(requestID, rc, req, 2);
    }

    private void executeInternal(String requestID, ResultCallback rc, String req, int count) {
        JsonObject jo2 = (new JsonParser()).parse(req).getAsJsonObject();
        String contractID = jo2.get("contractID").getAsString();
        // TODO 标注失效节点，是否选择重新迁移？？
        ResultCallback cb =
                new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        LOGGER.debug(str);
                        JsonObject jo = (new JsonParser()).parse(str).getAsJsonObject();
                        jo.remove("action");
                        jo.addProperty("action", "onExecuteResult");
                        LOGGER.debug(jo.toString());
                        if (jo.has("data")) {
                            String data = jo.get("data").getAsString();
                            ContractResult cr = JsonUtil.fromJson(data, ContractResult.class);
                            if (cr.status != ContractResult.Status.Success && count > 0) {
                                executeInternal(requestID, rc, req, count - 1);
                            } else
                                rc.onResult(jo.toString());
                        } else {
                            JsonObject jo2 = new JsonObject();
                            jo2.addProperty("action", "onExecuteResult");
                            jo.remove("action");
                            jo2.addProperty("data", jo.toString());
                            rc.onResult(jo2.toString());
                        }
                    }
                };
        NodeCenterActions.sync.sleepWithTimeout(requestID, cb, 5);
        if (!sendOnce(req))
            rc.onResult(
                    "{\"status\":\"Error\",\"data\":\"{\\\"status\\\":\\\"Error\\\",\\\"result\\\":\\\"all nodes offline\\\"}\",\"action\":\"onExecuteContract\"}");
    }

    private boolean sendOnce(String req) {
        JsonObject jo2 = (new JsonParser()).parse(req).getAsJsonObject();
        String contractID = jo2.get("contractID").getAsString();

        for (int i = 0; i < info.members.size(); i++) {
            int size = info.members.size();
            String nodeID = info.members.get(order.incrementAndGet() % size);
            CMNode cmNode = NodeCenterActions.nodeInfos.get(nodeID);

            if (cmNode != null) {
                cmNode.connection.controller.sendMsg(req);
                return true;
            }
        }
        return false;
    }
}
