package org.bdware.server.nodecenter;

import com.google.gson.*;
import com.google.gson.reflect.TypeToken;
import io.netty.util.Timeout;
import io.netty.util.TimerTask;
import io.netty.util.internal.StringUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractResult;
import org.bdware.sc.ContractResult.Status;
import org.bdware.sc.bean.ContractDesp;
import org.bdware.sc.bean.ContractExecType;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.encrypt.HardwareInfo;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.action.*;
import org.bdware.server.permission.Role;
import org.bouncycastle.pqc.math.linearalgebra.ByteUtils;
import org.zz.gmhelper.SM2Util;

import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

public class NodeCenterActions {
    // static SyncResult syncResult = new SyncResult();
    private static final Logger LOGGER = LogManager.getLogger(NodeCenterActions.class);
    private static final String MISSING_ARGUMENT = "Missing arguments";
    public static Map<String, CMNode> nodeInfos =
            new ConcurrentHashMap<>(); // all online nodes,key是节点pubKey

    public static SyncResult sync = new SyncResult();

    public static Map<String, MultiPointContractInfo> contractID2Members =
            new ConcurrentHashMap<>(); // contractID,list(nodes' pubKey)
    //    public static ConcurrentHashMap<String, ConcurrentHashMap<String, ContractRecord>>
    // recoverMap =
    //            new ConcurrentHashMap<>(); // k-nodesPubKey,v-(k-contractID),only unit contract
    // TODO  定期清缓存
    //    public static Map<String, RequestCache> requestCache =
    //            new ConcurrentHashMap<>(); // key is contractID,only for requestAll type contract
    static TimerTask checkAliveTask;
    static Timeout checkAliveTimeout;

    static {
        int delay = 1000;
        checkAliveTask =
                new TimerTask() {
                    @Override
                    public void run(Timeout arg0) {
                        try {
                            if (!arg0.isCancelled()) {
                                checkAlive();
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                        if (!arg0.isCancelled()) {
                            checkAliveTimeout =
                                    SyncResult.timer.newTimeout(this, delay, TimeUnit.MILLISECONDS);
                        }
                    }

                    // heartbeat?(zyx)
                    private void checkAlive() {
                        Set<String> toRemove = new HashSet<>();
                        for (String key : nodeInfos.keySet()) {
                            if (!nodeInfos.get(key).checkAlive()) {
                                LOGGER.info("Node " + key.substring(0, 5) + " is offline!");
                                //                                LOGGER.info(
                                //                                        "[NodeCenterActions] 发现节点
                                // "
                                //                                                + key.substring(0,
                                // 5)
                                //                                                + " not alive!");
                                toRemove.add(key);
                            }
                        }

                        for (String key : toRemove) {
                            CMNode info = nodeInfos.get(key);
                            LOGGER.debug("Heart beat: " + info.nodeName + " is offline!");
                            //                            System.out.println("[ADSP]NC心跳机制发现节点 " +
                            // info.nodeName + " 下线!");

                            synchronized (info) {
                                if (null != info.contracts && !info.contracts.isEmpty()) {
                                    for (ContractDesp cd : info.contracts) {
                                        cd.setIsMaster(false);
                                        LOGGER.info(
                                                "checkAlive---- 设置节点 "
                                                        + info.pubKey.substring(0, 5)
                                                        + " 的合约 "
                                                        + cd.contractID
                                                        + " isMaster="
                                                        + false);
                                    }
                                }
                            }

                            nodeInfos.remove(key);
                            try {
                                info.connection.controller.ctx.close();

                            } catch (Exception ignored) {
                            }

                            /*  for(String contractID : recoverMap.get(info.pubKey).keySet()){
                                MasterActions.unitModeCheck(contractID);
                                ContractRecord record = recoverMap.get(info.pubKey).get(contractID);
                                synchronized (record) {
                                    if (record.recoverFlag == RecoverFlag.Fine)
                                        record.recoverFlag = RecoverFlag.ToRecover;
                                }
                            }*/
                        }
                    }
                }; // checkAliveTask
        checkAliveTimeout =
                SyncResult.timer.newTimeout(checkAliveTask, delay, TimeUnit.MILLISECONDS);

        // checkAliveTimeout = SyncResult.timer.newTimeout(checkAliveTask, 20, TimeUnit.SECONDS);

        //        NodeCenterServer.scheduledThreadPool.scheduleWithFixedDelay(
        //                () -> {
        //                    boolean flag = clearCache();
        //                    if (flag) {
        //                        try {
        //                            Thread.sleep(10000L);
        //                        } catch (InterruptedException e) {
        //                            LOGGER.error("sleeping is interrupted! " + e.getMessage());
        //                        }
        //                    }
        //                },
        //                0,
        //                0,
        //                TimeUnit.SECONDS);
    }

    public NodeCenterFrameHandler controller;
    String nodeID; // node pubKey
    boolean NCConFlag = false; // true表示这个是别的NC连过来的
    Map<String, FileOutputStream> fileMap = new HashMap<>();
    private String sessionID;
    private String nodeManagerPubkey;

    public NodeCenterActions(NodeCenterFrameHandler nodeCenterFrameHandler) {
        controller = nodeCenterFrameHandler;
    }

    public static ContractDesp getContractByID(String key) {
        for (CMNode node : nodeInfos.values()) {
            for (ContractDesp desp : node.contracts) {
                if (desp.contractID.equals(key)) return desp;
            }
        }
        return null;
    }

    public static ContractDesp getContractByName(String key) {
        for (CMNode node : nodeInfos.values()) {
            for (ContractDesp desp : node.contracts) {
                if (desp.contractID.equals(key) || desp.contractName.equals(key)) return desp;
            }
        }
        return null;
    }

    public static String getPeerIdByNodeId(String id) {
        CMNode node = nodeInfos.get(id);
        return node.peerID;
    }

    public static String getIpPortByNodeId(String id) {
        CMNode node = nodeInfos.get(id);
        return node.ipPort;
    }

    @Action(httpAccess = true)
    public void downloadUUID(JsonObject args, ResultCallback resultCallback) {
        String pubKey = args.get("pubKey").getAsString();
        String signature = args.get("sign").getAsString();
        String nonce = args.get("nonce").getAsString();
        if (StringUtil.isNullOrEmpty(pubKey)
                || StringUtil.isNullOrEmpty(signature)
                || StringUtil.isNullOrEmpty(nonce)) {
            resultCallback.onResult("missing pubkey/sign/nonce");
        }

        if (NCManagerAction.isCenterManager(pubKey)) {

            boolean verify = SM2Util.plainStrVerify(pubKey, nonce, signature);
            if (!verify) {
                resultCallback.onResult("verify failed");
                return;
            }
            String uuid =
                    ByteUtil.encodeBASE64(HardwareInfo.getCPUID().getBytes()).replaceAll("\n", "");
            ((HttpResultCallback) resultCallback)
                    .addHeader("content-disposition", "attachment;filename=encodeduuid.key");
            resultCallback.onResult(uuid);
        } else {
            resultCallback.onResult("permission denied, not ContractManager");
        }
    }

    private Response listCMInfoInterval(JsonObject arg) {
        Response r = new Response();
        r.action = "onListCMInfo";
        r.data = new HashMap<>(nodeInfos);
        if (arg.has("requestID")) {
            r.responseID = arg.get("requestID").getAsString();
        }
        return r;
    }
    //
    //    @Action(userPermission = 0)
    //    public void login(JsonObject json, ResultCallback resultCallback) {
    //        long start = System.currentTimeMillis();
    //        if (sessionID == null)
    //            resultCallback.onResult("{\"action\":\"onLogin\",\"data\":\"failed\"}");
    //        String signature = json.get("signature").getAsString();
    //        String pubKey = json.get("pubKey").getAsString();
    //        SM2 sm2 = new SM2();
    //        boolean result =
    //                sm2.verify(
    //                        (sessionID),
    //                        Signature.loadFromString(signature),
    //                        "",
    //                        SM2KeyPair.publicKeyStr2ECPoint(pubKey));
    //        logger.debug("session:" + (sessionID));
    //        if (result) {
    //            this.nodeManagerPubkey = pubKey;
    //            logger.debug("设置公钥" + pubKey);
    //            resultCallback.onResult("{\"action\":\"onLogin\",\"data\":\"success\"}");
    //        } else {
    //            resultCallback.onResult("{\"action\":\"onLogin\",\"data\":\"failed\"}");
    //        }
    //        long l = getPermission(nodeManagerPubkey);
    //        controller.setPermission(l);
    //        long end = System.currentTimeMillis();
    //        logger.debug("time:" + (end - start));
    //    }

    private void sendContractResult(JsonObject json) {
        controller.sendMsg(json.toString());
    }

    @Action
    public void syncPing(JsonObject json, ResultCallback rc) {
        // logger.debug("[NodeCenterAction] syncPing");
        if (json.has("data")) {
            String data = json.get("data").getAsString();
            LOGGER.debug("[syncPing] data" + data);
            CMNode node;
            node = nodeInfos.get(json.get("id").getAsString());
            switch (data.charAt(0)) {
                case '+':
                    node.addCIManager(data.substring(1));
                    break;
                case '-':
                    node.removeCIManager(data.substring(1));
                    break;
                case '@': // peerID
                    node.peerID = data.substring(1);
                    break;
                default:
                    LOGGER.debug("[syncPing] wrong data" + data);
            }
            LOGGER.debug("[syncPing] info change");
        }
        if (json.has("contractVersion")) {
            CMNode node = nodeInfos.get(nodeID);
            // LOGGER.debug("[syncPing] contractVersion:"+json.get("contractVersion").getAsInt()+"
            // -- "+node.contractVersion);

            if (null != node && json.get("contractVersion").getAsInt() != node.contractVersion) {
                rc.onResult("{\"action\":\"syncContractInfo\"}");
            }
        }
        rc.onResult(
                "{\"action\":\"syncPong\",\"requestID\":\""
                        + json.get("requestID").getAsString()
                        + "\"}");
    }

    @Action(userPermission = 0)
    public void getSessionID(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        sessionID = new Random().nextLong() + "_session";
        Map<String, String> result = new HashMap<>();
        result.put("action", "onSessionID");
        result.put("session", sessionID);
        resultCallback.onResult(JsonUtil.toJson(result));
        long end = System.currentTimeMillis();
        LOGGER.debug("cmclient time:" + (end - start) + "data:" + JsonUtil.toJson(result));
    }

    private long getPermission(String pubKey) {
        if (pubKey == null) {
            return 0L;
        }
        try {
            String ret;
            ret = KeyValueDBUtil.instance.getValue(NCTables.NodeUser.toString(), pubKey);
            LOGGER.info("Role=" + ret + " --> pubkey=" + pubKey);
            if (null == ret || ret.isEmpty()) {
                /*
                KeyValueDBUtil.instance.setValue(
                        NCTables.NodeUser.toString(),
                        pubKey,
                        Role.Node.toString());
                KeyValueDBUtil.instance.setValue(
                        NCTables.NodeTime.toString(),
                        pubKey,
                        Long.toString(new Date().getTime()));
                return Role.Node.getValue();
                */
                // TODO fix permission bugs.
                return Role.compoundValue(new String[]{"NodeManager"});
            } else {

                return (Role.compoundValue(ret.split(",")));
            }
        } catch (Exception e) {
            e.printStackTrace();
            return 0L;
        }
    }

    @Action(userPermission = 0)
    public void setNodeID(JsonObject json, ResultCallback rc) {
        LOGGER.debug(json.toString());
        if (rc instanceof HttpResultCallback) {
            Response r = new Response();
            r.action = "setNodeID";
            r.data = "forbidden";
            rc.onResult(JsonUtil.toJson(r));
            return;
        }
        Response r = new Response();
        r.action = "onSetNodeID";
        String pubkey = json.get("id").getAsString();
        String signature = json.get("signature").getAsString();
        LOGGER.debug("id:" + pubkey + " sig:" + signature);
        r.data = "failed";
        if (sessionID != null && SM2Util.plainStrVerify(pubkey, pubkey + sessionID, signature)) {
            nodeID = pubkey;
            controller.pubKey = nodeID;
            CMNode node = new CMNode(this, pubkey);
            String nodeShortID = node.pubKey.substring(0, 5);
            if (json.has("cimanager")) {
                node.setCIManager(json.get("cimanager").getAsString());
            }
            if (json.has("peerID")) {
                node.setPeerID(json.get("peerID").getAsString());
            }
            if (json.has("ipPort")) {
                node.setIpPort(json.get("ipPort").getAsString());
            }
            long permission = getPermission(pubkey);
            if (permission != 0) {
                controller.setPermission(permission);
                LOGGER.info("node " + nodeShortID + " is online!");
                nodeInfos.put(pubkey, node);
                if (json.has("nodeName")) {
                    node.nodeName = json.get("nodeName").getAsString();
                    KeyValueDBUtil.instance.setValue(
                            NCTables.NodesDB.toString(), pubkey, node.nodeName);
                    LOGGER.info("set node name: " + nodeShortID + " -> " + node.nodeName);
                }
                if (json.has("masterAddress")) {
                    node.masterAddress = json.get("masterAddress").getAsString();
                }

                r.data = "success";

                if (null != KeyValueDBUtil.instance.getValue(NCTables.NodesDB.toString(), pubkey)) {
                    KeyValueDBUtil.instance.setValue(NCTables.NodesDB.toString(), pubkey, pubkey);
                }
                // Map<String, String> udpReq = new HashMap<>();
                // udpReq.put("action", "getUDPID");
                /// udpReq.put("udpPort", runner.port + "");
                // rc.onResult(JsonUtil.toJson(udpReq));
            }
        }

        // TODO 如果是NC挂了，其他节点重连之后需要另外考虑
        // 向该合约发送请求，令其查看自己崩溃前的集群合约
        LOGGER.info("NodeCenter tells nodes new masters for contracts created before the crash!");
        Map<String, String> request = new HashMap<>();
        request.put("action", "queryUnitContractsID");
        try {
            CMNode conn = nodeInfos.get(pubkey);
            NodeCenterActions connection = conn.connection;
            NodeCenterFrameHandler connController = connection.controller;
            connController.sendMsg(JsonUtil.toJson(request));
        } catch (Exception e) {
            e.printStackTrace();
        }
        Map<String, String> onNewNodeConnected = new HashMap<>();
        onNewNodeConnected.put("action", "onNewNodeConnected");
        onNewNodeConnected.put("pubKey", pubkey);
        for (CMNode node : nodeInfos.values()) {
            if (node.pubKey.equals(pubkey)) {
                continue;
            }
            try {
                node.connection.controller.sendMsg(JsonUtil.toJson(onNewNodeConnected));
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        rc.onResult(JsonUtil.toJson(r));
    }

    @Action(userPermission = 0)
    public void getNCConnectPermission(JsonObject json, ResultCallback rc) {
        LOGGER.debug("[NodeCenterAction] getNCConnectPermission " + json.toString());
        if (rc instanceof HttpResultCallback) {
            Response r = new Response();
            r.action = "onGetNCConnectPermission";
            r.data = "forbidden";
            rc.onResult(JsonUtil.toJson(r));
            return;
        }
        Response r = new Response();
        r.action = "onGetNCConnectPermission";
        String pubkey = json.get("id").getAsString();
        String signature = json.get("signature").getAsString();
        LOGGER.debug("[NodeCenterController] id:" + pubkey + " sig:" + signature);
        if (sessionID != null && SM2Util.plainStrVerify(pubkey, pubkey + sessionID, signature)) {
            r.data = "success";
            NCConFlag = true;
        } else r.data = "failed";

        rc.onResult(JsonUtil.toJson(r));
    }

    @Action(httpAccess = true)
    public void listCMInfo(JsonObject json, ResultCallback rc) {
        // 检查是否为在线节点
        //        if (!json.containsKey("pubKey") || !nodeinfos.containsKey(json.get("pubKey"))) {
        //            LOGGER.error("node pubKey not valid " + json.get("pubKey"));
        //            return;
        //        }
        Response r = listCMInfoInterval(json);
        rc.onResult(JsonUtil.toJson(r));
    }

    @Action(async = true, userPermission = 1)
    public void checkIsContract(JsonObject json, ResultCallback rc) {
        JsonArray array = json.get("list").getAsJsonArray();
        JsonObject jo = new JsonObject();
        jo.addProperty("action", "onCheckIsContract");
        JsonObject jo1 = new JsonObject();
        for (int i = 0; i < array.size(); i++) {
            String key = array.get(i).getAsString();
            String val = KeyValueDBUtil.instance.getValue(NCTables.ContractMeta.toString(), key);
            jo1.addProperty(key, val != null && val.length() > 0);
        }
        jo.addProperty("responseID", json.get("requestID").getAsString());
        jo.add("result", jo1);
        rc.onResult(jo.toString());
    }

    @Action(async = true, userPermission = 1L)
    public void updateContract(JsonObject json, ResultCallback rc) throws IOException {
        // TODO @fanbo 收到了更新请求时，检查本地数据库，看是否有contractID对应的索引。
        // 如果没有，就rc.onResult("{\"action\":\"requestReadMe\",\"contractID\":\"xxx\"}");
        CMNode node = nodeInfos.get(nodeID);
        if (null != node) {
            // 对与旧的master挂了，但是合约进程没挂，旧master重新上线后需要看现在该合约是否有新的master
            // 如果有就更改json串，并发消息给该节点将Contract的isMaster设置为false
            Set<String> contractIDS = new HashSet<>(); // 旧master需要自己非master的合约
            // 当前节点的合约
            List<ContractDesp> contracts =
                    JsonUtil.fromJson(
                            json.get("contracts"),
                            new TypeToken<List<ContractDesp>>() {
                            }.getType());
            MetaIndexAction.updateContractsIndex(contracts, rc);
            LOGGER.debug("update contracts: " + json.get("contracts"));
            int version = -1;
            if (json.has("contractVersion")) {
                version = json.get("contractVersion").getAsInt();
            }
            node.updateContract(contracts, version);
            if (json.has("events")) {
                node.events = json.get("events").getAsInt();
            }
            // 遍历所有节点
            for (CMNode cmNode : nodeInfos.values()) {
                if (cmNode.pubKey.equals(nodeID)) {
                    continue;
                }
                if (null == cmNode.contracts || cmNode.contracts.isEmpty()) {
                    continue;
                }
                // 遍历其他节点的所有合约
                for (ContractDesp desp : cmNode.contracts) {
                    if (desp.type == ContractExecType.Sole) {
                        continue;
                    }
                    // 遍历当前节点的所有合约
                    for (ContractDesp thisDesp : contracts) {
                        /// logger.debug("测试位置555 节点" + cmNode.pubKey.substring(0,5) + "
                        // 的合约 " + desp.contractID + " " + desp.isMaster + "  与 节点 " +
                        // nodeID.substring(0,5) + " 的合约 " + thisDesp.contractID + "  " +
                        // thisDesp.isMaster);

                        if (!((desp.contractID).equals(thisDesp.contractID))) {
                            continue;
                        }
                        if (thisDesp.type == ContractExecType.Sole) {
                            continue;
                        }

                        if (desp.getIsMaster() && thisDesp.getIsMaster()) {
                            LOGGER.info(
                                    "合约 "
                                            + thisDesp.contractID
                                            + "  的旧的master "
                                            + node.pubKey.substring(0, 5)
                                            + "重新上线后和现在master  "
                                            + cmNode.pubKey.substring(0, 5)
                                            + "撞车!");
                            contractIDS.add(thisDesp.contractID);
                        }
                    }
                }
            }

            StringBuilder con_ids = new StringBuilder();
            for (ContractDesp cd : contracts) {
                if (contractIDS.contains(cd.contractID)) {
                    cd.setIsMaster(false);
                    LOGGER.info(
                            "设置合约 "
                                    + cd.contractID
                                    + " 的旧master "
                                    + node.pubKey.substring(0, 5)
                                    + " isMaster为false!");
                    con_ids.append(",").append(cd.contractID);
                }
            }
            if (!con_ids.toString().equals("")) {
                Map<String, String> request = new HashMap<>();
                request.put("action", "updateNonMasters");
                request.put("contracts", con_ids.toString());
                node.connection.controller.sendMsg(JsonUtil.toJson(request));
            }
            if (rc instanceof HttpResultCallback) {
                rc.onResult("{\"msg\":\"success\"}");
            }
        }
    }

    @Action(async = true, userPermission = 1)
    public void checkAlive(JsonObject json, ResultCallback rc) {
        checkAliveTimeout.cancel();
        checkAliveTimeout = SyncResult.timer.newTimeout(checkAliveTask, 0, TimeUnit.SECONDS);
        rc.onResult("{\"action\":\"pong\",\"data\":" + System.currentTimeMillis() + "}");
    }

    @Action(async = true, userPermission = 1L << 27)
    public void takeScreenshot(JsonObject json, ResultCallback rc) {
        String nodeName = json.get("nodeName").getAsString();
        if (nodeName == null) {
            rc.onResult("{\"status\":\"error\",\"data\":\"missing argument: nodeName\"}");
            return;
        }
        for (CMNode node : nodeInfos.values()) {
            if (node.nodeName.equals(nodeName)
                    || node.pubKey.equals(nodeName)
                    || node.udpID.equals(nodeName)
                    || node.ipPort.equals(nodeName)) {
                String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
                Map<String, String> req = new HashMap<>();
                req.put("requestID", requestID);
                req.put("action", "takeScreenshot");
                node.connection.controller.sendMsg(JsonUtil.toJson(req));
                sync.sleepWithTimeout(requestID, rc, 20);
                return;
            }
        }
        rc.onResult("{\"status\":\"error\",\"data\":\"can't local device!\"}");
    }

    @Action(async = true)
    public void onTakeScreenshot(JsonObject json, ResultCallback rc) {
        sync.wakeUp(json.get("responseID").getAsString(), json.get("data").getAsString());
    }

    @Action(async = true, userPermission = 1)
    public void onReceiveContractExecution(JsonObject json, ResultCallback rc) {
        if (rc instanceof HttpResultCallback) {
            Response r = new Response();
            r.action = "onReceiveContractExecution";
            r.data = "forbidden";
            rc.onResult(JsonUtil.toJson(r));
            return;
        }
        String requesterID = json.get("requesterNodeID").getAsString();
        CMNode node = nodeInfos.get(requesterID);
        if (node != null) node.connection.sendContractResult(json);
        if (requesterID.equals("NodeCenter")) {
            sync.wakeUp(
                    json.get("requestID").getAsString(), json.get("contractResult").getAsString());
        }
    }

    @Action(async = true, userPermission = 1)
    public void executeContractOnOtherNodes(JsonObject json, ResultCallback rc) {
        if (rc instanceof HttpResultCallback) {
            Response r = new Response();
            r.action = "onExecuteContractOnOtherNodes";
            r.data = "forbidden";
            rc.onResult(JsonUtil.toJson(r));
            return;
        }
        LOGGER.debug(
                "tid:"
                        + Thread.currentThread().getId()
                        + "executeContractOnOtherNodes\n"
                        + JsonUtil.toJson(json));
        String requestID = json.get("requestID").getAsString();
        try {
            CMNode node;
            String crStr = json.get("contractRequest").getAsString();
            ContractRequest cr = JsonUtil.fromJson(crStr, ContractRequest.class);
            String requesterNodeID = nodeID;
            for (String nodeID : nodeInfos.keySet()) {
                if (nodeID.equals(this.nodeID)) continue;
                node = nodeInfos.get(nodeID);
                if (node != null && node.contracts != null)
                    for (ContractDesp cd : node.contracts) {
                        if (cr.getContractID().equals(cd.contractID)
                                || cr.getContractID().equals(cd.contractName)) {
                            JsonObject jo = new JsonObject();
                            jo.addProperty("requestID", requestID);
                            jo.addProperty("requesterNodeID", requesterNodeID);
                            jo.addProperty("action", "executeContractLocally");
                            jo.addProperty("contractRequest", crStr);
                            LOGGER.debug(
                                    "tid:"
                                            + Thread.currentThread().getId()
                                            + "execute in node:"
                                            + node.nodeName);
                            node.connection.controller.sendMsg(JsonUtil.toJson(jo));
                            return;
                        }
                    }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        LOGGER.debug("tid:" + Thread.currentThread().getId() + "can't locate in NodeCenter!");
        ContractResult cResult =
                new ContractResult(
                        Status.Error,
                        new JsonPrimitive("can't locate the Contract in node center"));
        JsonObject jo = new JsonObject();
        jo.addProperty("action", "onReceiveContractExecution");
        jo.addProperty("requestID", requestID);
        jo.addProperty("contractResult", JsonUtil.toJson(cResult));
        rc.onResult(JsonUtil.toJson(jo));
    }

    @Action
    public void getLogSize(JsonObject json, ResultCallback cb) {
        syncRequestCM(json, cb);
    }

    @Action
    public void requestLog(JsonObject json, ResultCallback cb) {
        syncRequestCM(json, cb);
    }

    @Action(userPermission = 2)
    public void requestLastLog(JsonObject json, ResultCallback cb) {
        syncRequestCM(json, cb);
    }

    private void syncRequestCM(JsonObject json, final ResultCallback cb) {
        String requestID = System.currentTimeMillis() + " | " + Math.random() * 1000;
        json.addProperty("requestID", requestID);
        Map<String, String> response = new HashMap<>();
        response.put("status", "Error");
        if (!json.has("contractID")) {
            response.put("result", "missing argument: contractID");
            cb.onResult(JsonUtil.toJson(response));
            return;
        }
        boolean hasContract = false;
        for (CMNode node : nodeInfos.values()) {
            if (node.containsContract(json.get("contractID").getAsString())) {
                hasContract = true;
                node.connection.controller.sendMsg(JsonUtil.toJson(json));
                break;
            }
        }
        if (!hasContract) {
            response.put("data", "can not locate contractID: " + json.get("contractID"));
            cb.onResult(JsonUtil.toJson(response));
        }
        sync.sleep(
                requestID,
                new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        ContractResult ret = JsonUtil.fromJson(str, ContractResult.class);
                        Map<String, Object> converted = new HashMap<>();
                        converted.put("status", ret.status);
                        try {
                            List<Map<String, String>> je =
                                    JsonUtil.fromJson(
                                            ret.result,
                                            new TypeToken<
                                                    List<Map<String, String>>>() {
                                            }.getType());
                            converted.put("result", je);
                        } catch (Exception e) {
                            converted.put("result", ret.result);
                        }
                        String retStr = JsonUtil.toJson(converted);
                        cb.onResult(retStr);
                    }
                });
    }

    /*  // judge the just online node whwther need to restart contracts,send the contracts'info
    public static void restartContracts(String nodePubKey) {
      if (!recoverMap.containsKey(nodePubKey)) {
        return;
      }

      // 恢复该节点的每一个集群运行的合约
      for (ContractRecord record : recoverMap.get(nodePubKey).values()) {
        String contractID = record.contractID;
        if (record.recoverFlag != RecoverFlag.ToRecover) continue;

        // 先发消息，让恢复节点的该合约收到消息后只加入队列不dealRequests
        Map<String, String> request = new HashMap<>();
        request.put("action", "setRecovering");
        request.put("contractID", contractID);
        CMNode node = nodeinfos.get(nodePubKey);
        node.connection.controller.sendMsg(gson.toJson(request));

        System.out.println(
            "第一步 : [NodeCeterActions] restartContracts  开始处理合约  contractID=" + contractID);

        if (contractID2Members.containsKey(contractID)) {

          // 在nodeinfos中找节点，该节点的pubKey在pubKeys中
          MultiPointContractInfo info = contractID2Members.get(contractID);
          CMNode cmNode = null;
          for (int i = 0; i < info.members.size(); i++) {
            int size = info.members.size();
            String tempNodeID = info.members.get(record.order.incrementAndGet() % size);

            if (nodeinfos.containsKey(tempNodeID))
              cmNode = NodeCenterActions.nodeinfos.get(tempNodeID);
            else continue;

            System.out.println("查询节点  " + cmNode.nodeName);

            if (cmNode != null && !cmNode.pubKey.equals(nodePubKey)) {
              System.out.println("第二步 : [NodeCenterActions] 找到一个依赖恢复节点，其节点名为 " +
               cmNode.nodeName);

              Map<String, String> req = new HashMap<String, String>();
              req.put("action", "dumpCurrentState");
              req.put("contractID", contractID);
              req.put("targetNodePubkey", nodePubKey);

              // NC向该节点发送请求，让其存储自身当前状态并发给NC
              cmNode.connection.controller.sendMsg(gson.toJson(req));

              return;
            }
          }

          if (cmNode == null) {
            logger.debug("[NodeCenterActions] Can't find a recover rely node!");
          }
        }
      }
    }*/

    // zyx modified
    /*    @Action(async = true)
    public void receiveState(Map<String, String> args, final ResultCallback rc) {
        //System.out.println("第五步  : [NodeCenterActions] 开始从恢复依赖节点接收state");

        String fileName = args.get("fileName") + "_NC";
        boolean isAppend = Boolean.valueOf(args.get("isAppend"));
        boolean isDone = Boolean.valueOf(args.get("isDone"));

        //logger.debug("[NodeCenterActions] isAppend=" + isAppend + "  isDone=" + isDone);

        String path = "./temp/stateFiles/" + fileName;
        File file = new File(path);
        File dir = file.getParentFile();
        if (!dir.exists()) {
            dir.mkdirs();
        }

        FileOutputStream fout = null;
        if (!isAppend) {
            try {
                fout = new FileOutputStream(file);
                fileMap.put(fileName, fout);
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            }
        } else {
            fout = fileMap.get(fileName);
        }

        if (isDone) {
            if (fout != null) {
                try {
                    fout.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            // TODO 参数可以加入文件名？
            //System.out.println("第七步  : [NodeCenterActions] 从恢复依赖节点接收state完成");

            //logger.debug("[NodeCenterActions] receive state finish.");
            String receiveFileName = new File(dir, fileName).getAbsolutePath();

            // 发送state给需要恢复的节点
            sendState(fileName, args.get("targetNodePubkey"));
        } else {
            String data = args.get("data");
            try {
                String progress = args.get("progress");
                // TODO pulish progress
                fout.write(ByteUtil.decodeBASE64(data));
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }*/

    // The resultCallback is an Counter, it will invoke
    // in its
    // onResult
    // zyx modified
    /*    public void sendState(String path, String targetNodePubkey) {
        File project = new File("./temp/stateFiles/" + path);

        // 将合约中的memberd加入ContractRecord
        ContractRecord record = null;
        try {
            FileInputStream fileout = new FileInputStream(project);
            GZIPInputStream gzin = new GZIPInputStream(fileout);
            ObjectInputStream reader = new ObjectInputStream(gzin);
            record = (ContractRecord) reader.readObject();
            reader.close();
        } catch (IOException | ClassNotFoundException e) {
            e.printStackTrace();
        }
        // record.members = contractID2MembersUDP.get(record.contractID);
        //NC delete state file
        if(project.isFile() && project.exists()){
            project.delete();
        }

        record.members = new HashMap<>(recoverMap.get(nodeID).get(record.contractID).members);

        File file = new File("./temp/stateFiles/" + path);
        File parent = file.getParentFile();
        if (!parent.exists()) parent.mkdirs();
        ObjectOutputStream writer;
        try {
            FileOutputStream fileout = new FileOutputStream(file);
            GZIPOutputStream out = new GZIPOutputStream(fileout);
            writer = new ObjectOutputStream(out);
            writer.writeObject(record);
            writer.flush();
            writer.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        //System.out.println("第八步  : [NodeCenterActions] 开始向需要恢复的节点转发State");
        //record.printContent();

        CMNode targetNode = nodeinfos.get(targetNodePubkey);

        Map<String, String> req = new HashMap<>();
        req.put("action", "receiveState");
        req.put("isAppend", "false");
        req.put("fileName", path);
        req.put("isDone", "false");
        String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
        req.put("requestID", requestID);

        if (project.isFile()) {
            try {
                FileInputStream fin = new FileInputStream(project);
                byte[] buff = new byte[30 * 1024];
                long count = 0;
                long total = project.length();

                for (int len = 0; (len = (fin.read(buff))) > 0; ) {
                    //logger.debug("len = " + len);
                    String data = ByteUtil.encodeBASE64(buff, len);
                    req.put("data", data);
                    count += len;
                    targetNode.connection.controller.sendMsg(gson.toJson(req));
                    req.put("isAppend", "true");

                    Thread.sleep(300);
                }
                fin.close();

                req.put("isDone", "true");
                req.remove("data");

                targetNode.connection.controller.sendMsg(gson.toJson(req));
                //NC delete state file
                if(file.isFile() && file.exists()){
                    file.delete();
                }

                //logger.debug("[NodeCenterActions] send state project finish.");

                //System.out.println("第十步  : [NodeCenterActions] 向需要恢复的节点转发State完成");
            } catch (IOException e) {
                e.printStackTrace();
            } catch (InterruptedException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }*/

    @Action
    public void onReceiveSyncRequest(JsonObject jo, final ResultCallback cb) {
        if (cb instanceof HttpResultCallback) {
            Response r = new Response();
            r.action = "pubEvent";
            r.data = "forbidden";
            cb.onResult(JsonUtil.toJson(r));
            return;
        }
        String data = jo.get("contractResult").getAsString();
        String requestID = jo.get("requestID").getAsString();
        sync.wakeUp(requestID, data);
    }

    // userPermission = 256
    @Action(async = true, httpAccess = true, userPermission = 0)
    public void executeContract(JsonObject args, final ResultCallback rc) {
        final long start = System.currentTimeMillis();
        ContractRequest cr = new ContractRequest();

        if (!args.has("contractID")) {
            rc.onResult(MISSING_ARGUMENT);
            return;
        }
        String contractID = args.get("contractID").getAsString();

        cr.setContractID(contractID);
        if (!args.has("arg")) {
            rc.onResult(MISSING_ARGUMENT);
            return;
        }
        if (args.has("operation")) {
            cr.setAction(args.get("operation").getAsString());
            cr.setArg(args.get("arg").getAsString());
        }

        final JsonObject ret = new JsonObject();

        ret.addProperty("action", "onExecuteResult");
        if (args.has("requester")) {
            cr.setPublicKey(args.get("requester").getAsString());
            byte[] sign = ByteUtils.fromHexString(args.get("signature").getAsString());
            try {
                sign = SM2Util.encodeSM2SignToDER(sign);
            } catch (Exception e) {
                e.printStackTrace();
            }
            cr.setSignature(ByteUtils.toHexString(sign));
        }
        if (args.has("pubkey")) {
            cr.setPublicKey(args.get("pubkey").getAsString());
            cr.setSignature(args.get("signature").getAsString());
        }
        String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
        String crStr = JsonUtil.toJson(cr);
        ret.addProperty("responseID", requestID);
        boolean hasResult = false;

        for (String nodeID : nodeInfos.keySet()) {
            if (nodeID.equals(this.nodeID)) continue;
            CMNode node = nodeInfos.get(nodeID);
            if (node != null && node.contracts != null)
                for (ContractDesp cd : node.contracts) {
                    if (cr.getContractID().equals(cd.contractID)
                            || cr.getContractID().equals(cd.contractName)) {
                        JsonObject jo = new JsonObject();
                        jo.addProperty("requestID", requestID);
                        jo.addProperty("requesterNodeID", "NodeCenter");
                        jo.addProperty("action", "executeContractLocally");
                        jo.addProperty("contractRequest", crStr);
                        node.connection.controller.sendMsg(JsonUtil.toJson(jo));
                        sync.sleep(
                                requestID,
                                new ResultCallback() {
                                    @Override
                                    public void onResult(String str) {
                                        ContractResult cr =
                                                JsonUtil.fromJson(str, ContractResult.class);
                                        ret.add("result", cr.result);
                                        ret.addProperty("status", cr.status + "");
                                        long executeTime = (System.currentTimeMillis() - start);
                                        ret.addProperty("executeTime", executeTime + "");
                                        String retStr = JsonUtil.toJson(ret);
                                        rc.onResult(retStr);
                                    }
                                });
                        hasResult = true;
                        break;
                    }
                }
        }
        if (!hasResult) {
            ContractResult result =
                    new ContractResult(
                            Status.Error, new JsonPrimitive("Can't locate contract At NodeCenter"));
            ret.addProperty("result", JsonUtil.toJson(result));
            long executeTime = (System.currentTimeMillis() - start);
            ret.addProperty("executeTime", executeTime + "");
            String retStr = JsonUtil.toJson(ret);
            rc.onResult(retStr);
        }
    }

    // zyx modified
    @Action(async = true)
    public void transmitRecords(JsonObject args, final ResultCallback rc) {
        String targetID = args.get("targetNodePubKey").getAsString();
        args.remove("targetNodePubKey");
        args.addProperty("action", "addRequestQueue");

        CMNode cmNode = NodeCenterActions.nodeInfos.get(targetID);
        cmNode.connection.controller.sendMsg(JsonUtil.toJson(args));
    }

    // -------------UDP--------
    @Action
    public void onGetUDPID(JsonObject args, final ResultCallback rc) {
        String udpID = args.get("udpID").getAsString();
        CMNode info = nodeInfos.get(nodeID);
        LOGGER.debug("NC  onGetUDPID  节点  " + info.nodeName + " 上线后设置udpID");

        info.udpID = udpID;
    }

    @Action
    public void listContractID2Members(JsonObject args, final ResultCallback rc) {
        JsonObject jo = new JsonObject();
        jo.addProperty("action", "onListContractID2Members");
        jo.addProperty("status", "Success");
        // jo.addProperty("data", JsonUtil.toJson(contractID2Members));
        rc.onResult(jo.toString());
    }

    @Action(async = true)
    public void receiveTrustfullyResult(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("tid" + Thread.currentThread().getId() + "reqID" + args.get("responseID"));
        sync.wakeUp(args.get("responseID").getAsString(), JsonUtil.toJson(args));
        // we don't need any response.
    }

    @Action(async = true)
    public void onStartContractTrustfully(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("[onStartContractTrustfully] " + args.get("requestID"));
        sync.wakeUp(args.get("requestID").getAsString(), JsonUtil.toJson(args));
        // we don't need any response.
    }

    // cb will invoke sendProject in the cluster in its onResult method.
    public void requestProject(
            String projectName, String pubkey, boolean isPrivate, ResultCallback cb) {
        LOGGER.debug("[NodeCenterActions] requestProject : -----------position2");

        Map<String, Object> req = new HashMap<>();
        req.put("action", "sendProject");
        req.put("isPrivate", isPrivate);
        req.put("projectName", projectName);
        String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
        req.put("requestID", requestID);
        req.put("pubKey", pubkey);
        // sync.wakeUp --> receiveProject done;

        sync.sleepWithTimeout(requestID, cb, 60);
        controller.sendMsg(JsonUtil.toJson(req));
    }

    // The resultCallback is an Counter, it will invoke
    // in its
    // onResult
    public void sendProject(
            String filePath, String isPrivate, String pubKey, String nodeIP, ResultCallback result) {
        LOGGER.debug("sendProject : position----6" + filePath);

        File project = new File(filePath);
        String projectName = project.getName().split("_NC")[0];
        LOGGER.debug("[sendProject] " + projectName);
        Map<String, String> req = new HashMap<>();
        req.put("action", "receiveProject");
        req.put("isPrivate", isPrivate);
        if (isPrivate.equals("true")) {
            ;
            req.put("pubKey", pubKey);
        }
        req.put("isAppend", "false");
        req.put("fileName", projectName);
        req.put("isDone", "false");
        String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
        req.put("requestID", requestID);
        // sync.wakeup at --> onReceiveProject
        sync.sleep(requestID, result);

        if (project.isFile()) {
            try {
                FileInputStream fin = new FileInputStream(project);
                byte[] buff = new byte[30 * 1024];
                long count = 0;
                long preCount = 0;
                long total = project.length();
                long step = total / 10;
                LOGGER.debug("ypk = " + total);

                for (int len = 0; (len = (fin.read(buff))) > 0; ) {
                    LOGGER.debug("len = " + len);
                    String data = ByteUtil.encodeBASE64(buff, len);
                    req.put("data", data);
                    count += len;
                    controller.sendMsg(JsonUtil.toJson(req));
                    req.put("isAppend", "true");
                    if (count - preCount > step || count == total) {
                        preCount = count;
                        Map<String, String> req2 = new HashMap<>();
                        req2.put("operation", "onDistribute");
                        req2.put("requestID", requestID);
                        req2.put("progress", String.format("%.2f", count * 100F / total));
                        req2.put("nodeIP", nodeIP);
                        sync.wakeUp(requestID, JsonUtil.toJson(req2));
                    }
                }
                fin.close();

                req.put("isDone", "true");
                req.remove("data");
                controller.sendMsg(JsonUtil.toJson(req));

                // delete ypk_NC
                LOGGER.debug("[NodeCenterActions] send project finish.");
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    @Action(async = true)
    public void onReceiveProject(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("[NodeCenterActions] onReceiveProject : position----8");

        // TODO maybe add pubkey into args.
        Map<String, String> req = new HashMap<>();
        req.put("operation", "onReceive");
        req.put("peerID", NodeCenterActions.getPeerIdByNodeId(args.get("nodeID").getAsString()));
        req.put("ipPort", NodeCenterActions.getIpPortByNodeId(args.get("nodeID").getAsString()));
        req.put("progress", args.get("progress").getAsString());
        req.put("requestID", args.get("requestID").getAsString());
        String requestID = args.get("requestID").getAsString();
        sync.wakeUp(requestID, JsonUtil.toJson(req));
    }

    /*
    @Action(async = true)
    public void startContractInEachNode(Map<String, String> args,final ResultCallback rc) {
    	logger.debug("[NodeCenterActions] startContractInEachNode : ");

    	Map<String, Object> req = new HashMap<>();
    	req.put("action", "startContractByYpk");
    	req.put("fileName", args.get("fileName"));
    	req.put("contractOwner",args.get("pubKey"));
    	req.put("contractSig",args.get("signature"));

    	String requestID = System.currentTimeMillis() + "_" + (int) (Math.random() * 10000);
    	req.put("requestID", requestID);
    	sync.sleep(requestID, rc);
    	controller.sendMsg(gson.toJson(req));
    }


    @Action(async = true)
    public void onStartEachNode(Map<String, String> args, final ResultCallback rc) {
    	logger.debug("[NodeCenterActions] onStartEachNode : ");

    	logger.debug("---action" + args.get("action"));
    	logger.debug("---requestID" + args.get("requestID"));
    	logger.debug("---data" + args.get("data"));
    	logger.debug("---cid" + args.get("cid"));
    	logger.debug("---executeTime" + args.get("executeTime"));


    	String str = "onStartEachNode;" + args.get("data") + ";" + args.get("cid") + ";" +
    	  args.get("executeTime");
    	sync.wakeUp(args.get("requestID"),str);
    }
    */

    @Action(async = true)
    public void receiveProject(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("[NodeCenterActions] receiveProject : position----4");

        String fileName = args.get("fileName").getAsString() + "_NC";
        boolean isAppend = args.get("isAppend").getAsBoolean();
        boolean isDone = args.get("isDone").getAsBoolean();
        boolean isPrivate = args.get("isPrivate").getAsBoolean();

        LOGGER.debug("[NodeCenterActions] isAppend=" + isAppend + "  isDone=" + isDone);

        String path = "./temp/";
        File dir = new File(path);
        if (!dir.exists()) {
            dir.mkdirs();
        }

        FileOutputStream fout = null;
        if (!isAppend) {
            try {
                fout = new FileOutputStream(new File(dir, fileName));
                fileMap.put(fileName, fout);
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            }
        } else {
            fout = fileMap.get(fileName);
        }

        if (isDone) {
            if (fout != null) {
                try {
                    fout.close();
                    fileMap.remove(fileName);
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            // TODO 参数可以加入文件名？
            LOGGER.debug("[NodeCenterActions] receive finish. position----4");
            String receiveFileName = new File(dir, fileName).getAbsolutePath();

            Map<String, String> req = new HashMap<>();
            req.put("operation", "distribute");
            req.put("requestID", args.get("requestID").getAsString());
            req.put("isPrivate", isPrivate + "");
            req.put("receiveFileName", receiveFileName);
            sync.wakeUp(args.get("requestID").getAsString(), JsonUtil.toJson(req));
        } else {
            String data = args.get("data").getAsString();
            try {
                String progress = args.get("progress").getAsString();
                byte[] encodedData = ByteUtil.decodeBASE64(data);
                assert null != fout;
                assert null != encodedData;
                // TODO pulish progress
                fout.write(encodedData);

                Map<String, String> req = new HashMap<>();
                req.put("operation", "NCreceive");
                req.put("progress", progress);
                req.put("requestID", args.get("requestID").getAsString());
                sync.wakeUp(args.get("requestID").getAsString(), JsonUtil.toJson(req));

            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    @Action(async = true)
    public void requestConnectToMaster(JsonObject args, final ResultCallback rc) {
        String master = args.get("master").getAsString();
        Map<String, String> key2address = new HashMap<>();
        // nodes' pubKey
        JsonArray members = args.get("members").getAsJsonArray();
        Map<String, Object> request = new HashMap<>();
        request.put("action", "requestConnectToMaster");
        request.put("members", members); // nodes' pubKey
        request.put("master", master);
        if (args.has("connectAll"))
            request.put("connectAll", args.get("connectAll").getAsBoolean());
        if (args.has("contractID")) {
            request.put("contractID", args.get("contractID"));
        }
        request.put("key2address", key2address); // nodes' pubKey , master
        for (int i = 0; i < members.size(); i++) {
            String str = members.get(i).getAsString();
            if (nodeInfos.containsKey(str)) {
                CMNode node = nodeInfos.get(str);
                key2address.put(str, node.masterAddress);
            }
        }
        String req = JsonUtil.toJson(request);
        for (int i = 0; i < members.size(); i++) {
            String str = members.get(i).getAsString();
            if (nodeInfos.containsKey(str)) {
                CMNode node = nodeInfos.get(str);
                node.connection.controller.sendMsg(req);
            }
        }
    }

    //    static boolean clearCache() {
    //
    //        if (requestCache.isEmpty()) return true;
    //
    //        final long time = System.currentTimeMillis() - 120000L; // 120s
    //        requestCache
    //                .entrySet()
    //                .removeIf(
    //                        entry -> {
    //                            RequestCache cache = entry.getValue();
    //                            if (cache == null) return true;
    //                            return cache.getTime() < time;
    //                        });
    //
    //        return false;
    //    }

    //    static RequestCache getCache(String contractID) {
    //        if (contractID != null) {
    //            RequestCache cache = requestCache.get(contractID);
    //            if (cache != null) return cache;
    //            else {
    //                logger.debug("[NodeCenterActions] create requestcache:" + contractID);
    //                RequestCache reqc = new RequestCache();
    //                requestCache.put(contractID, reqc);
    //                return reqc;
    //            }
    //        }
    //        return null;
    //    }

    /*    @Action(async = true)
    public void sendCachedRequests(Map<String, String> args, final ResultCallback rc) {
        String contractID = args.get("contractID");
        int start = Integer.parseInt(args.get("start"));
        int end = Integer.parseInt(args.get("end"));

        RequestCache cache = getCache(contractID);
        for (int i = start + 1; i < end; i++) {
            if (cache == null || !cache.containsKey(i)) {

                // this node crash
                String nodeID = args.get("nodeID");
                restartContracts(nodeID);
            }
            JsonObject jo = cache.get(i);
            controller.sendMsg(JsonUtil.toJson(jo));

            logger.debug("NC发送第 " + jo.get("seq").getAsString() + " 个请求给Node");
        }
    }*/

    /*    @Action(async = true)
    public void recoverFinish(Map<String, String> args, final ResultCallback rc) {
        ContractRecord cr = recoverMap.get(args.get("nodeID")).get(args.get("contractID"));
        if (cr.recoverFlag == RecoverFlag.ToRecover) cr.recoverFlag = RecoverFlag.Fine;
    }*/

    @Action(async = true)
    public void electMaster(JsonObject args, final ResultCallback rc) {
        String contractID = args.get("contractID").getAsString();
        int lastExe = args.get("lastExe").getAsInt();
        String nodeID = args.get("nodeID").getAsString();

        LOGGER.info("electMaster nodeID=" + nodeID);

        ElectMasterTimeRecorder.nodeFindMasterCrash.put(nodeID, System.currentTimeMillis());

        String masterID = args.get("master").getAsString(); // 旧的master
        String members = args.get("members").getAsString(); // 执行这个合约所有节点的pubkey
        String uniNumber = null;
        if (args.has("uniNumber")) uniNumber = args.get("uniNumber").getAsString();

        //        if (nodeinfos.containsKey(masterID)) {
        //            CMNode node = nodeinfos.get(masterID);
        //            synchronized (node) {
        //                for (ContractDesp cd : node.contracts) {
        //                    if (cd.contractID.equals(contractID) ||
        // cd.contractName.equals(contractID)) {
        //                        cd.setIsMaster(false);
        //                        LOGGER.debug(
        //                                "选举步骤1---- 设置节点 "
        //                                        + node.pubKey.substring(0, 5)
        //                                        + " 的合约 "
        //                                        + contractID
        //                                        + " isMaster="
        //                                        + false);
        //                    }
        //                }
        //            }
        //        }

        synchronized (NCElectMasterUtil.electInfos) {
            if (!NCElectMasterUtil.electInfos.containsKey(contractID)) {
                NCElectMasterUtil.electInfos.put(
                        contractID,
                        new NCElectMasterUtil.ElectInfo(masterID, contractID, members, uniNumber));
            }
            NCElectMasterUtil.ElectInfo eleInfo = NCElectMasterUtil.electInfos.get(contractID);
            eleInfo.put(nodeID, lastExe, masterID, members, uniNumber);
        }
    }

    @Action(async = true)
    public void queryAllRouteInfo(JsonObject args, final ResultCallback rc) {
        args.addProperty("action", "onQueryAllRouteInfo");
        JsonArray array = new JsonArray();
        // JsonObject
        for (Map.Entry<String, CMNode> stringCMNodeEntry : nodeInfos.entrySet()) {
            CMNode node = stringCMNodeEntry.getValue();
            if (node.contracts == null) continue;
            for (ContractDesp desp : node.contracts) {
                if (desp.type == ContractExecType.Sole || desp.getIsMaster()) {
                    array.add(extractContractRoute(node, desp));
                }
            }
        }
        args.add("result", array);
        rc.onResult(args.toString());
    }

    private JsonObject extractContractRoute(CMNode node, ContractDesp desp) {
        JsonObject info = new JsonObject();
        info.addProperty("pubKey", node.pubKey);
        info.addProperty("masterAddress", node.masterAddress);
        info.addProperty("contractID", desp.contractID);
        info.addProperty("contractName", desp.contractName);
        return info;
    }

    @Action(async = true)
    public void queryNodeAddress(JsonObject args, final ResultCallback rc) {
        args.addProperty("action", "onQueryNodeAddress");
        args.add("responseID", args.get("requestID"));
        args.add("result", JsonNull.INSTANCE);
        String nodePubkey = args.get("pubKey").getAsString();
        CMNode nodeInfo = nodeInfos.get(nodePubkey);
        if (nodeInfo != null) {
            JsonObject jo = new JsonObject();
            jo.addProperty("pubKey", nodePubkey);
            jo.addProperty("masterAddress", nodeInfo.masterAddress);
            args.add("result", jo);
        }
        rc.onResult(args);
    }

    @Action(async = true)
    public void queryRouteInfo(JsonObject args, final ResultCallback rc) {
        String contractID = args.get("contractID").getAsString();
        args.addProperty("action", "onQueryRouteInfo");
        args.add("responseID", args.get("requestID"));
        args.add("result", JsonNull.INSTANCE);
        if (contractID == null || contractID.length() == 0) {
            args.add("result", new JsonPrimitive("Not Found"));
        }
        for (Map.Entry<String, CMNode> stringCMNodeEntry : nodeInfos.entrySet()) {
            CMNode node = stringCMNodeEntry.getValue();
            if (node.contracts == null) continue;
            for (ContractDesp desp : node.contracts) {
                if (contractID.equals(desp.contractID) || contractID.equals(desp.contractName)) {
                    LOGGER.info(
                            "查看合约 "
                                    + desp.contractID
                                    + "  "
                                    + desp.contractName
                                    + " 节点 "
                                    + node.pubKey.substring(0, 5)
                                    + " "
                                    + desp.type
                                    + "  "
                                    + desp.getIsMaster());
                    if (desp.type == ContractExecType.Sole) {
                        args.add("result", extractContractRoute(node, desp));
                    } else if (desp.getIsMaster()) {
                        LOGGER.info(
                                "合约 " + contractID + " 的master为 " + node.pubKey.substring(0, 5));
                        args.add("result", extractContractRoute(node, desp));
                    }
                }
            }
        }

        /*        logger.debug("查看合约 " + contractID + " 的master为 " + args.get("result"));
        if(!args.containsKey("result") || args.get("result") == null || args.get("result").equals("null")){
            //NC发起各个节点重选
            logger.debug("NC发现合约 " + contractID + " 的master为null，NC向该合约的各个节点发起重选请求!");
            Map<String,String> req = new HashMap<>();
            req.put("action","NCStartElect");
            req.put("contractID",contractID);
            for(CMNode node : nodeinfos.values()){
                node.connection.controller.sendMsg(JsonUtil.toJson(req));
                break;
            }
        }*/

        if (args.get("result").equals("null")) {
            LOGGER.info("[NodeCenterActions] query route by info from other NC");

            // 查询其他节点的信息
            String s = OtherNCProxy.instance.search(contractID);
            LOGGER.info(
                    "[NodeCenterActions] query route by info from other NC "
                            + (s == null ? "null" : s));
            if (s != null) {
                String[] ss = s.split(";");
                args.addProperty(
                        "result",
                        "{\"pubKey\":\"" + ss[0] + "\",\"masterAddress\":\"" + ss[1] + "\"}");
            }
        }

        rc.onResult(JsonUtil.toJson(args));
    }

    @Action(async = true)
    public void NCStartElect(JsonObject args, final ResultCallback rc) {
        String contractID = args.get("contractID").getAsString();

        LOGGER.debug("查看合约 " + contractID + " 的master为 " + args.get("result"));
        if (!args.has("result")
                || args.get("result") == null
                || args.get("result").getAsString().equals("null")) {
            // NC发起各个节点重选
            LOGGER.debug("NC发现合约 " + contractID + " 的master为null，NC向该合约的各个节点发起重选请求!");
            Map<String, String> req = new HashMap<>();
            req.put("action", "NCStartElect");
            req.put("contractID", contractID);
            req.put("uniNumber", System.currentTimeMillis() + "_" + new Random().nextLong());
            for (CMNode node : nodeInfos.values()) {
                node.connection.controller.sendMsg(JsonUtil.toJson(req));
                LOGGER.info(
                        "[ADSP]  [NodeCenterActions] 发现合约 "
                                + contractID
                                + " 的master为null 向节点 "
                                + node.nodeName
                                + "发送选举请求");
            }
        }
    }

    // 另可能仅和旧的master断开和NC连接的节点连新的master
    //    @Action(async = true)
    //    public void updateFormerMaster(JsonObject args, final ResultCallback rc) {
    //        String[] offlines = args.get("offlines").getAsString().split(",");
    //        String contractID = args.get("contractID").getAsString();
    //        String master = args.get("master").getAsString();
    //
    //        for (String nodeID : offlines) {
    //            if (nodeinfos.containsKey(nodeID)) {
    //                CMNode node = nodeinfos.get(nodeID);
    //                for (ContractDesp desp : node.contracts) {
    //                    if (desp.contractID.equals(contractID)
    //                            || desp.contractName.equals(contractID)) {
    //                        Map<String, String> request = new HashMap<>();
    //                        request.put("action", "setClientNotMaster");
    //                        request.put("contractID", contractID);
    //                        node.connection.controller.sendMsg(JsonUtil.toJson(request));
    //                        request.put("action", "queryUnitContractsID");
    //                        node.connection.controller.sendMsg(JsonUtil.toJson(request));
    //                    }
    //                }
    //            }
    //        }
    //    }

    @Action(async = true)
    public void changeMaster(JsonObject args, final ResultCallback rc) {
        String contractID = args.get("contractID").getAsString();
        String master = args.get("master").getAsString();
        LOGGER.debug("设置合约 " + contractID + " 的master为 " + master.substring(0, 5));
        CMNode node = nodeInfos.get(master);
        synchronized (node) {
            for (ContractDesp cd : node.contracts) {
                if (cd.contractID.equals(contractID) || cd.contractName.equals(contractID)) {
                    cd.setIsMaster(true);
                    LOGGER.debug(
                            "设置节点 "
                                    + node.pubKey.substring(0, 5)
                                    + " 的合约 "
                                    + cd.contractID
                                    + "   "
                                    + cd.contractName
                                    + " isMaster="
                                    + cd.getIsMaster());
                }
            }
        }

        ElectMasterTimeRecorder.changeMasterFinish = System.currentTimeMillis();
    }

    @Action(async = true, userPermission = 0)
    public void distributeContract(JsonObject args, final ResultCallback rc) {
        LOGGER.debug("[UnitActions] distributeContract :  -----------position1");

        String signature = args.get("signature").getAsString();
        String projectName = args.get("projectName").getAsString();
        boolean isPrivate = args.get("isPrivate").getAsBoolean();
        String pubKey = args.get("pubkey").getAsString(); // 用户的
        LOGGER.debug("pubkey=" + pubKey);
        // String sponsorPubKey = args.get("sponsorPubKey").getAsString();
        String sponsorPubkey = args.get("sponsorPubkey").getAsString();
        LOGGER.debug("sponsorPubkey=" + sponsorPubkey);
        String[] strs =
                args.get("nodeIDs").getAsString().split(","); // all nodes' pubKey in the unit
        Set<String> nodePubKeys = new HashSet<>(); // nodes' pubkey
        for (String str : strs) {
            if (str != null && str.length() > 0) nodePubKeys.add(str);
        }
        Map<String, String> nodes = new HashMap<>(); // node's pubKey-node'a name
        for (CMNode node : NodeCenterActions.nodeInfos.values()) {
            if (nodePubKeys.contains(node.pubKey)) {
                nodes.put(node.pubKey, node.nodeName);
                LOGGER.debug("nodes add " + node.pubKey + "   " + node.nodeName);
            }
        }

        String distributeID = args.get("distributeID").getAsString();
        LOGGER.debug("distributeID=" + distributeID);
        DistributeCallback dc =
                new DistributeCallback(distributeID, sponsorPubkey, nodes, rc, pubKey, signature);

        LOGGER.debug("[UnitActions] signature=" + signature);
        LOGGER.debug("[UnitActions] pubKey=" + pubKey);

        // 身份验证

        boolean result =
                SM2Util.plainStrVerify(
                        pubKey, "DistributeContract|" + projectName + "|" + pubKey, signature);

        LOGGER.debug("[UnitAcitons] 验证:" + result + " -> projectName:" + projectName);

        // requestProject
        if (result) {
            for (CMNode node : NodeCenterActions.nodeInfos.values()) {
                LOGGER.debug(node.nodeName + "  " + node.pubKey);
            }
            CMNode sponsor = NodeCenterActions.nodeInfos.get(sponsorPubkey);
            LOGGER.debug("[UnitActions] sponsor=" + sponsor.pubKey);

            sponsor.connection.requestProject(projectName, pubKey, isPrivate, dc);
        } else {
            rc.onResult(
                    "{\"status\":\"Identity error\"" + ",\"action\":\"onStartTrustfulContract\"}");
        }
    }

    @Action
    public void requestRouteInfo(JsonObject args, final ResultCallback rc) {
        Map<String, String> to = new HashMap<>();
        List<?> info = OtherNCProxy.instance.getInfo();
        if (info.isEmpty()) {
            LOGGER.info("[NodeCenterActions] requestRouteInfo info is empty");
            to.put("empty", "true");
        }
        final String centerManger = "__CenterManager__";
        String id = KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), centerManger);
        to.put("action", "receiveUpdate");
        to.put("address", args.get("address").getAsString());
        to.put("id", id);
        to.put("contracts", JsonUtil.toJson(info));
        LOGGER.info(JsonUtil.toJson(info));

        controller.sendMsg(JsonUtil.toJson(to));
    }


    @Action(async = true)
    public void askElectMasterTimeRecorder(JsonObject args, ResultCallback resultCallback) {
        String data = "null";
        String item = args.get("item").getAsString();
        if (args.has("slave")) {
            String nodeID = args.get("nodeID").getAsString();
            if ("nodeFindMasterCrash".equals(item)) {
                if (ElectMasterTimeRecorder.nodeFindMasterCrash.containsKey(nodeID)) {
                    data = ElectMasterTimeRecorder.nodeFindMasterCrash.get(nodeID) + "";
                }
            } else {
                data = "null";
            }
        } else {
            switch (item) {
                case "startElect":
                    if (ElectMasterTimeRecorder.startElect != null) {
                        data = ElectMasterTimeRecorder.startElect + "";
                    }
                    break;
                case "findNewMaster":
                    if (ElectMasterTimeRecorder.findNewMaster != null) {
                        data = ElectMasterTimeRecorder.findNewMaster + "";
                    }
                    break;
                case "newMaster":
                    if (ElectMasterTimeRecorder.newMaster != null) {
                        data = ElectMasterTimeRecorder.newMaster + "";
                    }
                    break;
                case "changeMasterFinish":
                    if (ElectMasterTimeRecorder.changeMasterFinish != null) {
                        data = ElectMasterTimeRecorder.changeMasterFinish + "";
                    }
                    break;
                default:
                    data = "null";
                    break;
            }
        }

        JsonObject ret = new JsonObject();
        ret.addProperty("action", "onAskElectMasterTimeRecorder");
        ret.addProperty("data", data);
        resultCallback.onResult(ret.toString());
    }

    @Action(userPermission = 0, async = true)
    public void queryAEState(JsonObject json, ResultCallback resultCallback) {
        JsonObject ret = new JsonObject();
        ret.addProperty("action", "onQueryAEState");
        if (json.has("caller")) {
            ret.add(
                    "result",
                    JsonParser.parseString(
                            JsonUtil.toJson(
                                    ActionExecutor.getStatistic(
                                            json.get("caller").getAsString()))));
        } else {
            ret.add("result", JsonParser.parseString(JsonUtil.toJson(ActionExecutor.getAllData())));
        }
        resultCallback.onResult(ret.toString());
    }

    @Action(userPermission = 0, async = true)
    public void queryAECaller(JsonObject json, ResultCallback resultCallback) {
        JsonObject ret = new JsonObject();
        ret.addProperty("action", "onQueryAECaller");
        ret.add("result", JsonParser.parseString(JsonUtil.toJson(ActionExecutor.getAECallerSet())));
        resultCallback.onResult(ret.toString());
    }

    public static class ResultCollector extends ResultCallback {
        private final ResultCallback committer;
        List<String> result = new ArrayList<>();
        String requestID;
        int repeatCount;

        public ResultCollector(String reqID, ResultCallback committer, int count) {
            this.committer = committer;
            repeatCount = count;
            requestID = reqID;
        }

        // 这个最好是把commit的结果也放过去？不要集齐全部再返回，而是一项项返回？
        // 需要在服务端记录"不正常节点？"
        @Override
        public void onResult(String str) {
            result.add(str);
            repeatCount--;
            if (repeatCount > 0) {
                sync.sleepWithTimeout(requestID, this, 20);
            } else {
                committer.onResult(JsonUtil.toJson(result));
            }
        }
    }
}
