package org.bdware.bdserver;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonObject;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.server.action.Action;
import org.bdware.server.action.ActionExecutor;
import org.bdware.server.nodecenter.NCHttpHandler;
import org.bdware.server.nodecenter.NodeCenterWSFrameHandler;
import org.bdware.server.permission.Role;
import org.junit.Test;

import java.lang.reflect.Method;
import java.util.*;

public class PermissionHelper {
    static class Line {
        String action;
        transient long permission;
        String roles;
    }

    @Test
    public void listClusterHttpAction() {
        ActionExecutor<ResultCallback,   JsonObject> ae = NCHttpHandler.getActionExecutor();
        Map<String, ActionExecutor.Pair<Method, Object>> handlers = ae.getHandlers();
        EnumSet<Role> set = EnumSet.allOf(Role.class);
        set.remove(Role.ContractInstanceManager);
        set.remove(Role.ContractProvider);
        set.remove(Role.ContractUser);
        printActions(handlers, set);
    }

    @Test
    public void listClusterWSAction() {
        NodeCenterWSFrameHandler handler = new NodeCenterWSFrameHandler();
        Map<String, ActionExecutor.Pair<Method, Object>> handlers = handler.getAE().getHandlers();
        EnumSet<Role> set = EnumSet.allOf(Role.class);
        set.remove(Role.ContractInstanceManager);
        set.remove(Role.ContractProvider);
        set.remove(Role.ContractUser);

        printActions(handlers, set);
    }

    private void printActions(
            Map<String, ActionExecutor.Pair<Method, Object>> handlers, EnumSet<Role> set) {
        List<Line> lines = new ArrayList<>();
        for (String str : handlers.keySet()) {
            Method m = handlers.get(str).first();
            Action a = m.getAnnotation(Action.class);
            Line l = new Line();
            lines.add(l);
            l.action = str;
            l.permission = a.userPermission();
            l.roles = "";
            for (Role r : set)
                if ((r.getValue() & l.permission) == l.permission) l.roles += r.name() + ";";
            if (l.roles.equals("CenterManager;NodeManager;Node;Anonymous;")) l.roles = "任意角色";
        }
        lines.sort(
                new Comparator<Line>() {
                    @Override
                    public int compare(Line o1, Line o2) {
                        return o1.action.compareTo(o2.action);
                    }
                });
        lines.sort(
                new Comparator<Line>() {
                    @Override
                    public int compare(Line o1, Line o2) {
                        return o1.roles.compareTo(o2.roles);
                    }
                });

        Gson g = new GsonBuilder().setPrettyPrinting().create();
        for (Line l : lines) {
            System.out.println("| " + l.action + " |  | " + l.roles + " |");
        }
    }
}
